//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: Sat Jul 19 12:09:12 PDT 2003
// Last Modified: Tue Jul 22 21:01:22 PDT 2003
// Filename:      ...sig/examples/all/proll.cpp
// Web Address:   http://sig.sapp.org/examples/museinfo/humdrum/chor2scor.cpp
// Syntax:        C++; museinfo
//
// Description:   Convert a Humdrum chorale file into a SCORE data file.
//

#include "humdrum.h"

#include <string.h>
#include <ctype.h>


// function declarations
void   checkOptions      (Options& opts, int argc, char* argv[]);
void   example           (void);
void   usage             (const char* command);
void   convertToScore    (HumdrumFile& infile);
void   printNote         (const char* data, double hpos, HumdrumFile& infile,
                          int line, int staff, int direction, int voice);
void   printStaffInfo    (HumdrumFile& infile);
int    noteheadtype      (double duration);
void   printMeasure      (HumdrumFile& infile, int line, double hpos);
void   printHarm         (const char* text, double hpos, HumdrumFile& infile, 
                          int harmspine);
void   printTrailer      (void);
int    getdots           (double duration, int voice);
int    getflags          (double duration);
double makeaccidental    (HumdrumFile& infile, int line, int voice, 
                          Array<int>& lkey);
void   resetkey          (void);
int    peekAccidental    (HumdrumFile& infile, int line, int voice, 
                          Array<int>& lkey);
double getHpos           (HumdrumFile& infile, int line);
void   generateBeams     (HumdrumFile& infile, int voice);
void   checkForBeam      (HumdrumFile& infile, int voice, Array<int>& index, 
                          int i);
double getBeamStart      (double vpos1, double vpos2, int dir);
double getBeamEnd        (double vpos1, double vpos2, int dir);
void   fillnindex        (HumdrumFile& infile, Array<Array<int> >& nindex);
void   initoffset        (Array<Array<double> >& offset, 
                          Array<Array<int> >& nindex);
void   storeoffset       (double offset, int voice, int line);
double getoffset         (const char* lower, const char* upper);
void   printTitle        (const char* title);
void   printHarmKeys     (HumdrumFile& infile);
void   printKeyInfo      (const char* key, double hpos, double vpos, 
                          double absbeat);

// global variables
Options   options;       // database for command-line arguments
int       debugQ    = 0; // used with --debug option

Array<double>         tie(4);  // for keeping track of ties
Array<int>            key(7);  // for keeping track of key
Array<Array<int> >    vkey(4);  // for keeping track of accidentals
Array<Array<double> > offsets(4);  // for keeping track of beam offsets
Array<Array<int> >    nindex(4);  // for keeping track of beam offsets

///////////////////////////////////////////////////////////////////////////

int main(int argc, char* argv[]) {
   checkOptions(options, argc, argv);

   HumdrumFile infile;
   char* filenameIn  = options.getArg(1);
   infile.read(filenameIn);

   tie.setAll(0.0);
   key.setAll(0);
   int i;
   for (i=0; i<3; i++) {
      vkey[i].setSize(7);
      vkey[i].setAll(0);
   }

   fillnindex(infile, nindex);
   initoffset(offsets, nindex);
   convertToScore(infile);

   return 0;
}


///////////////////////////////////////////////////////////////////////////




//////////////////////////////
//
// printTitle --
//

void printTitle(const char* title) {
   int i;
   int length = strlen(title);

   for (i=0; i<length; i++) {
      if ((title[i] == 'a') && (title[i+1] == 'e')) {
         cout << "%%a";
         i++;
      } else if ((title[i] == 'o') && (title[i+1] == 'e')) {
         cout << "%%o";
         i++;
      } else if ((title[i] == 'u') && (title[i+1] == 'e')) {
         cout << "%%u";
         i++;
      } else {
         cout << title[i];
      }
   }
}



//////////////////////////////
//
// initoffset --
//

void initoffset(Array<Array<double> >& offsets, Array<Array<int> >& nindex) {
   int i;
   offsets.setSize(4);
   for (i=0; i<4; i++) {
      offsets[i].setSize(nindex[i].getSize());
      offsets[i].setAll(0.0);
   }
}



//////////////////////////////
//
// fillnindex --  create an index of the notes in each voice
//     for beam maintenance.
//

void fillnindex(HumdrumFile& infile, Array<Array<int> >& nindex) {
   int i, j;
   for (i=0; i<4; i++) {
      nindex[i].setSize(infile.getNumLines());
      nindex[i].setSize(0);
      for (j=0; j<infile.getNumLines(); j++) {
         if (infile[j].isData()) {
            if (strcmp(infile[j][i], ".") == 0) {
               continue;
            }
            if (strchr(infile[j][i], 'r') != NULL) {
               continue;
            }
            nindex[i].append(j);
         }
      }
   }
}



//////////////////////////////
//
// resetkey -- copy the global key into the temporary voice key for
//     the measure.
//

void resetkey(void) {
   int i;
   int j;
   for (i=0; i<4; i++) {
      for (j=0; j<7; j++) {
         vkey[i][j] = key[j];
      }
   }
}



//////////////////////////////
//
// setKey -- set the default key signature accidentals
//

void setKey(int keynum) {
   if (keynum < 0) {
      if (keynum <= -7) key[3] = -1; // F
      if (keynum <= -6) key[0] = -1; // C
      if (keynum <= -5) key[4] = -1; // G
      if (keynum <= -4) key[1] = -1; // D
      if (keynum <= -3) key[5] = -1; // A
      if (keynum <= -2) key[2] = -1; // E
      if (keynum <= -1) key[6] = -1; // B
   } else if (keynum > 0) {
      if (keynum >= 1) key[3] = 1; // F
      if (keynum >= 2) key[0] = 1; // C
      if (keynum >= 3) key[4] = 1; // G
      if (keynum >= 4) key[1] = 1; // D
      if (keynum >= 5) key[5] = 1; // A
      if (keynum >= 6) key[2] = 1; // E
      if (keynum >= 7) key[6] = 1; // B
   }
}



//////////////////////////////
//
// printStaffInfo -- print staff, keysignature, Title, and meter
//
// 50% p4 = 2:-36 3:-48 4:-84
// 40% p4 = 2:-45 3:-67 4:-112
//

void printStaffInfo(HumdrumFile& infile) {
   // print bass staff:
   cout << "8 1 0 0 0.4 200\n";
   // print tenor staff:
   cout << "8 2 0 -45 0.4 200\n";
   // print alto staff:
   cout << "8 3 0 -67 0.4 200\n";
   // print treble staff:
   cout << "8 4 0 -112 0.4 200\n";

   // print bass clef:
   cout << "3 1 2 0 1\n";
   // print bass clef:
   cout << "3 2 2 0 1\n";
   // print treble clef:
   cout << "3 3 2 0 0\n";
   // print treble clef:
   cout << "3 4 2 0 0\n";

   // print opening barlines
   cout << "14 1 0 4 0\n";
   cout << "14 1 0 4 8\n";

   // print key signatures 
   int i;
   int keynum = 0;
   for (i=0; i<infile.getNumLines(); i++) {
      if (infile[i].isInterpretation()) {
         if (strncmp(infile[i][0], "*k[", 3) == 0) {
            keynum = Convert::kernKeyToNumber(infile[i][0]);
         }
      }
   }
   setKey(keynum);
   vkey[0] = key;
   vkey[1] = key;
   vkey[2] = key;
   vkey[3] = key;

   if (keynum != 0) {
      // print bass key signature
      cout << "17 1 4 0 " << keynum << " 1\n";
      // print bass key signature
      cout << "17 2 4 0 " << keynum << " 1\n";
      // print treble key signature
      cout << "17 3 4 0 " << keynum << " 0\n";
      // print treble key signature
      cout << "17 4 4 0 " << keynum << " 0\n";
   } 

   // print time signatures 
   int top = 5;
   int bottom = 5;
   for (i=0; i<infile.getNumLines(); i++) {
      if (infile[i].isInterpretation()) {
         if (sscanf(infile[i][0], "*M%d/%d", &top, &bottom) == 2) {
            break;
         }
      }
   }

   int metc = 0;
   for (i=0; i<infile.getNumLines(); i++) {
      if (infile[i].isInterpretation()) {
         if (strcmp(infile[i][0], "*met(C)") == 0) {
            metc = 1;
            break;
         }
      }

   }

   if (metc) {
      // print the bass time signature
      cout << "18 1 7 0 " << 99 << " " << 1 << "\n";
      // print the bass time signature
      cout << "18 2 7 0 " << 99 << " " << 1 << "\n";
      // print the treble time signature
      cout << "18 3 7 0 " << 99 << " " << 1 << "\n";
      // print the treble time signature
      cout << "18 4 7 0 " << 99 << " " << 1 << "\n";
   } else {
      // print the bass time signature
      cout << "18 1 7 0 " << top << " " << bottom << "\n";
      // print the bass time signature
      cout << "18 2 7 0 " << top << " " << bottom << "\n";
      // print the treble time signature
      cout << "18 3 7 0 " << top << " " << bottom << "\n";
      // print the treble time signature
      cout << "18 4 7 0 " << top << " " << bottom << "\n";
   }

   // print title information
   char* title = "";
   for (i=0; i<infile.getNumLines(); i++) {
      if (infile[i].isBibliographic()) {
         if (strncmp(infile[i][0], "!!!OTL", 6) == 0) {
            if (strchr(infile[i][0], ':') != NULL) {
               title = strchr(infile[i][0], ':') + 1;
               while (title[0] == ' ' || title[0] == '\n') {
                  if (title[0] == '\0') {
                     break;
                  }
                  title++;
               }
               break;  // only get the first title in the file
            }
         }
      }
   }
   int erknum = 0;
   for (i=0; i<infile.getNumLines(); i++) {
      if (infile[i].isBibliographic()) {
         if (strncmp(infile[i][0], "!!!SCT: Erk", 11) == 0) {
            sscanf(infile[i][0], "!!!SCT: Erk %d", &erknum);
         }
      }
   }

   cout << "t 4 0 20 0 1.5\n";
   if (erknum > 0) {
      cout << "_00" << erknum << ".   ";
      printTitle(title);
      cout << "\n";
   } else {
      cout << "_00";
      printTitle(title);
      cout << "\n";
   }

   // print final barline
   cout << "14 1 200 4 2\n";
}



//////////////////////////////
//
// convertToScore -- convert humdrum data into a SCORE file
//

void convertToScore(HumdrumFile& infile) {
   int i;
   double hpos = 0.0;
   infile.analyzeRhythm();

   printStaffInfo(infile);

   int harmspine = 5;
   int staff = 0;
   int direction = 0;
   int voice = 0;
   for (i=0; i<infile.getNumLines(); i++) {
      hpos = getHpos(infile, i);
      if (infile[i].isMeasure()) {
         printMeasure(infile, i, hpos);
      } else if (infile[i].isData()) {
         // cout << infile[i].getAbsBeat() << ":\t" << hpos << endl;

         // check bass note
         if (strcmp(infile[i][0], ".") != 0) {
            staff = 1;
            voice = 0;
            direction = 20;
            printNote(infile[i][0], hpos, infile, i, 
                  staff, direction, voice);
         }
         // check tenor note
         if (strcmp(infile[i][1], ".") != 0) {
            staff = 2;
            voice = 1;
            direction = 10;
            printNote(infile[i][1], hpos, infile, i, 
                  staff, direction, voice);
         }
         // check alto note
         if (strcmp(infile[i][2], ".") != 0) {
            staff = 3;
            voice = 2;
            direction = 20;
            printNote(infile[i][2], hpos, infile, i, 
                  staff, direction, voice);
         }
         // check soprano note
         if (strcmp(infile[i][3], ".") != 0) {
            staff = 4;
            voice = 3;
            direction = 10;
            printNote(infile[i][3], hpos, infile, i, 
                  staff, direction, voice);
         }

         // print **harm data
         harmspine = 5;
         if (strcmp(infile[i][harmspine], ".") != 0) {
            printHarm(infile[i][harmspine], hpos, infile, harmspine);
         }

      }
   }

   for (i=0; i<4; i++) {
      generateBeams(infile, i);
   }

   printHarmKeys(infile);

   printTrailer();
}



//////////////////////////////
//
// printHarmKeys -- print the analysis key
//

void printHarmKeys(HumdrumFile& infile) {
   int i;
   int length;
   for (i=0; i<infile.getNumLines(); i++) {
      if (infile[i].isInterpretation()) {
         length = strlen(infile[i][0]);
         if ((infile[i][0][length-1] == ':') && 
             (Convert::kernToBase40(infile[i][0]) > 0)) {
            printKeyInfo(infile[i][0], getHpos(infile, i), -9, 
               infile[i].getAbsBeat());
         }
      }
   }
}



//////////////////////////////
//
// printKeyInfo -- print the analysis key
//

void printKeyInfo(const char* key, double hpos, double vpos, double absbeat) {
   int i;
   int length = strlen(key);
   double offset = length <= 3 ? -4 : -4.6;
   if (absbeat > 0) {
      offset += 1.75;
   }

   cout << "t 1 " << hpos << " " << vpos << " 0 1.0 1.25 0 0 0 "
        << offset
        << "\n";
   cout << "_01" ;


   for (i=1; i<length-1; i++) {
      if (key[i] == '-') {
         cout << "?1";
      } else if (key[i] == '#') {
         cout << "?2";
      } else {
         cout << key[i]; 
      }
   }
   cout << ":\n";
}



//////////////////////////////
//
// printTrailer --
//

void printTrailer(void) {
   cout << 
      "STM 99\n"
//      "IF p1=6 and p7>19 and (p4<8 or p5<8) then p4=p4+0.5, p5=p5+0.5\n"
//      "IF p1=6 and p7>19 and (p4<6 or p5<6) then p4=p4+0.5, p5=p5+0.5\n"
//      "IF p1=6 and p7>19 and (p4<4 or p5<4) then p4=p4+0.5, p5=p5+0.5\n"
      "AD 99\n"
      "LJ\n"
      "1 4\n\n\n"
   ;

}



//////////////////////////////
//
// printHarm -- print Harmonic analysis
//

void printHarm(const char* text, double hpos, HumdrumFile& infile, 
      int harmspine) {
   if (strchr(text, 'r') != NULL) {
      return;
   }

   cout << "t 1 " << hpos << " -7 0 0 0 0 90\n";
   cout << "_29" ;
   int i;
   int length = strlen(text);
   for (i=0; i<length; i++) {
      if (text[i] == '/') {
         cout << '\\';
      } else if (text[i] == ';') {
         // don't print fermata
      } else {
         cout << text[i];
      }
   }
   cout << "\n";
}



//////////////////////////////
//
// printMeasure --
//

void printMeasure(HumdrumFile& infile, int line, double hpos) {
   int btype;
   const char* data = infile[line][0];

   if (strchr(data, '-') != NULL) {
      return;  // this is a non-printing barline 
   } else if (strncmp(data, "==", 2) == 0) {
      return;  // don't print because header does this one
   } else if (strstr(data, ":|!|:") != NULL) {
      btype = 5;
   } else if (strstr(data, ":!!:") != NULL) {
      btype = 6;
   } else if (strstr(data, ":|!") != NULL) {
      btype = 3;
   } else if (strstr(data, "!|:") != NULL) {
      btype = 4;
   } else if (strstr(data, "||") != NULL) {
      btype = 1;
   } else {
      btype = 0;
   }

   cout << "14 1 " << hpos-0.25 << " 4 " << btype << "\n";

   // print measure number, if it is there
   int measureno = 0;
   if (sscanf(data, "=%d", &measureno)) {
      cout << "10 4 " << hpos-0.25 << " 13 " << measureno << " 0.8 1\n";
      cout << "I\nCM\n\n";
   }

   resetkey();
}



//////////////////////////////
//
// getoffset --
//

double getoffset(const char* lower, const char* upper) {
   if (strcmp(lower, ".") == 0) {
      return 0;
   }
   if (strcmp(upper, ".") == 0) {
      return 0;
   }

   int lowp = Convert::kernToBase40(lower);
   int upp  = Convert::kernToBase40(upper);
   if ((upp - lowp > 0) && (upp - lowp < 10)) {
      return 10;
   }

   int lowtype = noteheadtype(Convert::kernToDuration(lower));
   int uptype  = noteheadtype(Convert::kernToDuration(upper));
   
   if ((upp - lowp >= 0) && (upp - lowp < 5)) {
      if (lowtype != uptype) {
         return 10;
      }
   }

   return 0;
}



//////////////////////////////
//
// makeaccidental --
//

double makeaccidental(HumdrumFile& infile, int line, int voice, 
   Array<int>& lkey) {

   int acc = 0;
   int length = strlen(infile[line][voice]);
   int i;
   for (i=0; i<length; i++) {
      if (infile[line][voice][i] == '-') {
         acc += -1;
      } else if (infile[line][voice][i] == '#') {
         acc += 1;
      }
   }

   int base40   = Convert::kernToBase40(infile[line][voice]);
   int diatonic = Convert::base40ToDiatonic(base40);

   // don't write an accidental which is in the current key set
   if ((acc == lkey[diatonic]) && 
       (strchr(infile[line][voice], 'n') == NULL) &&
       (strchr(infile[line][voice], 'y') == NULL) 
      ) {
      return 0;
   }

   lkey[diatonic] = acc;
   int output = 0;
  
   if (strchr(infile[line][voice], 'n') != NULL) {
      output = 3;  // natural sign
   } else {
      switch (acc) {
         case -2: output = 4; break; // double flat
         case -1: output = 1; break; // flat
         case 0:  output = 3; break; // natural sign
         case 1:  output = 2; break; // sharp sign
         case 2:  output = 5; break; // double sharp sign
      }
   }

   // check for moving the accidental if they will collide

   double fraction = 0.0;
   if ((voice == 1) || (voice == 3)) {
      return output + fraction;
   } 

   // check on bass and alto voices

   int poutput = 0;
   if (voice == 0) {
      poutput = peekAccidental(infile, line, 1, vkey[1]);  // look at tenor acc.
   } else {
      poutput = peekAccidental(infile, line, 3, vkey[3]);  // look at sop. acc.
   }

   if (poutput == 0) {
      return output + fraction;
   }

   int b40u = 0;
   int b40d = 0;

   if (voice == 0) {
      b40u = Convert::kernToBase40(infile[line][1]);
      b40d = Convert::kernToBase40(infile[line][0]);
   } else {
      b40u = Convert::kernToBase40(infile[line][3]);
      b40d = Convert::kernToBase40(infile[line][2]);
   }

   // the function does not check properly if the voices are inverted
   if ((b40u - b40d < 0) || (b40u - b40d >= 40)) {
      return output + fraction; 
   }
   
   int dup = Convert::base40ToDiatonic(b40u);
   int ddn = Convert::base40ToDiatonic(b40d);

   int dint = dup - ddn;
   if (dint < 0) {
      dint += 7;
   }

   //  Here are the accidental offset separations by diatonic interval for
   //  sharps, flats and naturals (doubles to be done later, if ever)
   //
   //   bot top    2    3    4    5    6    7
   //    b   b    0.24 0.24 0.15 0.15 0.12 0.00
   //    b   s    0.25 0.25 0.25 0.20 0.18 0.11
   //    b   n    0.22 0.22 0.18 0.12 0.12 0.00
   //    s   b    0.24 0.24 0.24 0.18 0.00 0.00
   //    s   s    0.29 0.29 0.29 0.22 0.20 0.00
   //    s   n    0.24 0.24 0.23 0.11 0.11 0.00
   //    n   b    0.24 0.24 0.20 0.13 0.00 0.00
   //    n   s    0.26 0.26 0.25 0.20 0.12 0.00
   //    n   n    0.22 0.22 0.16 0.12 0.00 0.00
   //

   // ignoring spacing when dealing with double sharps or double flats
   if ((poutput > 3) || (output > 3)) {
      return output + fraction;
   }

   // ignoring spacing for unison accidentals:
   if (dint == 0) {
      return output + fraction;
   }

   switch (output * 10 + poutput) {
      case 11: 
         //    b   b    0.24 0.24 0.15 0.15 0.12 0.00
         switch (dint) { 
            case 1: case 2: fraction = 0.24; break;
            case 3: case 4: fraction = 0.15; break;
            case 5: fraction = 0.12; break; 
            case 6: fraction = 0.12; break; 
         }
         break;
      case 12:
         //    b   s    0.25 0.25 0.25 0.20 0.18 0.11
         switch (dint) { 
            case 1: case 2: case 3: fraction = 0.25; break;
            case 4: fraction = 0.20; break;
            case 5: fraction = 0.18; break;
            case 6: fraction = 0.11; break;
         }
         break;
      case 13:
         //    b   n    0.22 0.22 0.18 0.12 0.12 0.00
         switch (dint) { 
            case 1: case 2: fraction = 0.22; break;
            case 3: fraction = 0.18; break;
            case 4: case 5: fraction = 0.12; break;
         }
         break;
      case 21:
         //    s   b    0.24 0.24 0.24 0.18 0.00 0.00
         switch (dint) { 
            case 1: case 2: case 3: fraction = 0.24; break;
            case 4: fraction = 0.18; break;
         }
         break;
      case 22:
         //    s   s    0.29 0.29 0.29 0.22 0.20 0.00
         switch (dint) { 
            case 1: case 2: case 3: fraction = 0.29; break;
            case 4: fraction = 0.22; break;
            case 5: fraction = 0.20; break;
         }
         break;
      case 23:
         //    s   n    0.24 0.24 0.23 0.11 0.11 0.00
         switch (dint) { 
            case 1: case 2: fraction = 0.24; break;
            case 3: fraction = 0.22; break;
            case 4: case 5: fraction = 0.11; break;
         }
         break;
      case 31:
         //    n   b    0.24 0.24 0.20 0.13 0.00 0.00
         switch (dint) { 
            case 1: case 2: fraction = 0.24; break;
            case 3: fraction = 0.20; break;
            case 4: fraction = 0.13; break;
         }
         break;
      case 32:
         //    n   s    0.26 0.26 0.25 0.20 0.12 0.00
         switch (dint) { 
            case 1: case 2: fraction = 0.26; break;
            case 3: fraction = 0.25; break;
            case 4: fraction = 0.20; break;
            case 5: fraction = 0.12; break;
         }
         break;
      case 33:
         //    n   n    0.22 0.22 0.16 0.12 0.00 0.00
         switch (dint) { 
            case 1: case 2: fraction = 0.22; break;
            case 3: fraction = 0.16; break;
            case 4: fraction = 0.12; break;
         }
         break;
   }

   return output + fraction;
}



//////////////////////////////
//
// peekAccidental --
//

int peekAccidental(HumdrumFile& infile, int line, int voice, 
   Array<int>& lkey) {

   if (strcmp(infile[line][voice], ".") == 0) {
      return 0;
   }

   int acc = 0;
   int length = strlen(infile[line][voice]);
   int i;
   for (i=0; i<length; i++) {
      if (infile[line][voice][i] == '-') {
         acc += -1;
      } else if (infile[line][voice][i] == '#') {
         acc += 1;
      }
   }

   int base40   = Convert::kernToBase40(infile[line][voice]);
   int diatonic = Convert::base40ToDiatonic(base40);

   // don't write an accidental which is in the current key set
   if ((acc == lkey[diatonic]) && (strchr(infile[line][voice], 'n') == NULL)) {
      return 0;
   }

   int output = 0;
  
   if (strchr(infile[line][voice], 'n') != NULL) {
      output = 3;  // natural sign
   } else {
      switch (acc) {
         case -2: output = 4; break; // double flat
         case -1: output = 1; break; // flat
         case 0:  output = 3; break; // natural sign
         case 1:  output = 2; break; // sharp sign
         case 2:  output = 5; break; // double sharp sign
      }
   }

   return output;
}



//////////////////////////////
//
// printNote --
//

void printNote(const char* data, double hpos, HumdrumFile& infile,
      int line, int staff, int direction, int voice) {
   double duration = Convert::kernToDuration(data);
   int pitch = Convert::kernToBase40(data);
   int vpos = 0;
   int clef = 0;
   double offset = 0.0;
   double accidental = 0.0;
   switch (staff) {
      case 1: clef = 1; break;
      case 2: clef = 1; break;
      case 3: clef = 0; break;
      case 4: clef = 0; break;
   }
   int tiedir = 0;
   switch (voice) {
      case 0: tiedir = -1; break;
      case 1: tiedir = 1;  break;
      case 2: tiedir = -1; break;
      case 3: tiedir = 1;  break;
   }
   if (strchr(data, 'r') != NULL) {
      switch (voice) {
         case 0: case 2:
            cout << "2 " << staff << " " << hpos << " -4 0 0 " 
                 << duration << endl;
            break;
         case 1: case 3:
            cout << "2 " << staff << " " << hpos << " 4 0 0 " 
                 << duration << endl;
            break;
      }
   } else {
      accidental = makeaccidental(infile, line, voice, vkey[voice]);
      offset = 0.0;
      if (voice == 2) {
         offset = getoffset(infile[line][2], infile[line][3]);
         storeoffset(offset, voice, line);
      } else if (voice == 0) {
         offset = getoffset(infile[line][0], infile[line][1]);
         storeoffset(offset, voice, line);

      } else if (voice == 1) {
         // if the tenor is below the bass,
         // then offset the tenor by a small amount so that note
         // heads do not collide with the other voices stem.
         if (Convert::kernToBase40(infile[line][1]) < 
             Convert::kernToBase40(infile[line][0])) {
            offset = 0.45;
            storeoffset(offset, voice, line);
         }
      } else if (voice == 3) {
         // if the soprano is below the alto 
         // then offset the soprano by a small amount so that note
         // heads do not collide with the other voices stem.
         if (Convert::kernToBase40(infile[line][3]) < 
             Convert::kernToBase40(infile[line][2])) {
            offset = 0.45;
            storeoffset(offset, voice, line);
         }
      }


      double stemadj = 0.0;
      double fermataadj = 0.0;
      vpos = Convert::base40ToScoreVPos(pitch, clef);
      if ((strchr(data, ';') != NULL) && (voice == 0)) {
         if (vpos < 2) {
            stemadj = -3;  // LJ erases this data anyway
         } 
         if (vpos < 0) {
            fermataadj = 0.5;
         }
      }
      cout << "1 " << staff << " " << hpos << " " << vpos << " "
           << direction + accidental << " " << noteheadtype(duration) << " " 
           << duration << " " << stemadj << " " 
           << 10 * getdots(duration, voice) + getflags(duration)
           << " " << offset;
      if (strchr(data, ';') != NULL) {
         if (voice == 0) {
            cout << " -14";
            if (fermataadj != 0.0) {
               cout << " " << fermataadj;
            }
         } else if (voice == 3) {
            cout << " 14";
         } else if (voice == 1) {
            if (strchr(infile[line][0], ';') == NULL) {
               cout << " 14";
            }
         } else if (voice == 2) {
            if (strchr(infile[line][3], ';') == NULL) {
               cout << " -14";
            }
         }
      }
      cout << "\n";

      // check tie information

      if (voice == 0 || voice == 2) {
         if (vpos % 2) {
            vpos = vpos - 1;
         } else {
            vpos = vpos - 2;
         }
      } else {
         if (vpos % 2) {
            vpos = vpos + 1;
         } else {
            vpos = vpos + 2;
         }
      }

      if (offset == 10) {
         hpos += 1;
      }

      double roffset = 0.0;
      if (voice == 1 || voice == 3) {
         roffset = 0.25;
      }

      if (strchr(data, '[') != NULL) {
         tie[voice] = hpos;
      } else if (strchr(data, '_') != NULL) {
         cout << "5 " << staff << " " << tie[voice] - 0.25 << " "
              << vpos << " " << vpos << " " << hpos + roffset << " "
              << tiedir << " -2\nI\nCR\n\n";
         tie[voice] = hpos;
      } else if (strchr(data, ']') != NULL) {
         cout << "5 " << staff << " " << tie[voice] - 0.25 << " "
              << vpos << " " << vpos << " " << hpos + roffset << " "
              << tiedir << " -2\nI\nCR\n\n";
         tie[voice] = 0.0;
      } else {
         tie[voice] = 0.0;
      }
   }
}



//////////////////////////////
//
// storeoffset --
//

void storeoffset(double offset, int voice, int line) {
   int index = -1;
   int i;
   for (i=0; i<nindex[voice].getSize(); i++) {
      if (nindex[voice][i] == line) {
         index = i;
      }
   }

   if (index >= 0) {
      offsets[voice][index] = offset;
   }
}



//////////////////////////////
//
// getflags --
//

int getflags(double duration) {
   if (duration <= 1.0/128.0) {
      return 8;
   } else if (duration < 1.0/64.0) {
      return 7;
   } else if (duration < 1.0/32.0) {
      return 6;
   } else if (duration < 1.0/16.0) {
      return 5;
   } else if (duration < 1.0/8.0) {
      return 4;
   } else if (duration < 1.0/4.0) {
      return 3;
   } else if (duration < 1.0/2.0) {
      return 2;
   } else if (duration < 1.0/1.0) {
      return 1;
   } else { 
      return 0;
   }
}



//////////////////////////////
//
// getdots -- return the number of rhythm augmentation dots
//

int getdots(double duration, int voice) {
   char buffer[128] = {0};
   Convert::durationToKernRhythm(buffer, duration);
   int i = 0;
   int count = 0;
   while (buffer[i] != '\0') {
      if (buffer[i] == '.') {
         count++;
      }
      i++;
   }

   if ((count > 0) && (voice == 0 || voice == 2)) {
      count += 10; // 10 multiplied later
   }

   return count;
}



//////////////////////////////
//
// noteheadtype --
//

int noteheadtype(double duration) {
   if (duration < 1.9) {
      return 0;
   } else if (duration < 3.9) {
      return 1;
   } else if (duration < 7.9) {
      return 2;
   } else {
      return 3; 
   }
}



//////////////////////////////
//
// checkOptions -- validate and process command-line options.
//

void checkOptions(Options& opts, int argc, char* argv[]) {
   opts.define("debug=b",          "trace input parsing");   
   opts.define("author=b",         "author of the program");   
   opts.define("version=b",        "compilation information"); 
   opts.define("example=b",        "example usage"); 
   opts.define("help=b",           "short description"); 
   opts.process(argc, argv);
   
   // handle basic options:
   if (opts.getBoolean("author")) {
      cout << "Written by Craig Stuart Sapp, "
           << "craig@ccrma.stanford.edu, July 2003" << endl;
      exit(0);
   } else if (opts.getBoolean("version")) {
      cout << argv[0] << ", version: July 2003" << endl;
      cout << "compiled: " << __DATE__ << endl;
      cout << MUSEINFO_VERSION << endl;
      exit(0);
   } else if (opts.getBoolean("help")) {
      usage(opts.getCommand());
      exit(0);
   } else if (opts.getBoolean("example")) {
      example();
      exit(0);
   }


   debugQ    =  opts.getBoolean("debug");
}



//////////////////////////////
//
// example -- example usage of the program
//

void example(void) {
   cout <<
   "                                                                        \n"
   << endl;
}



//////////////////////////////
//
// usage -- gives the usage statement for the program
//

void usage(const char* command) {
   cout <<
   "                                                                        \n"
   << endl;
}



//////////////////////////////
//
// generateBeams -- 
//

void generateBeams(HumdrumFile& infile, int voice) {
   int i;
   Array<int> index;
   index.setSize(infile.getNumLines());
   index.setSize(0);
   for (i=0; i<infile.getNumLines(); i++) {
      if (infile[i].isData()) {
         if (strcmp(infile[i][voice], ".") == 0) {
            continue;
         }
         if (strchr(infile[i][voice], 'r') != NULL) {
            continue;
         }
         index.append(i);
      }
   }

   for (i=0; i<index.getSize(); i++) {
      checkForBeam(infile, voice, index, i); 
   }
}



//////////////////////////////
//
// checkForBeam -- see if a beam is supposed to start here
//

void checkForBeam(HumdrumFile& infile, int voice, Array<int>& index, int ind) {


   int startbeat = 0;
   int endbeat = 0;

   // assume that there are no pickups less than a quarter note.
   int beat = (int)(infile[index[ind]].getAbsBeat()+ 0.001);
   double fraction = infile[index[ind]].getAbsBeat() - beat;


   int tbeat = 0;
   int j;
   for (j=ind; j<index.getSize(); j++) {
      tbeat = (int)(infile[index[j]].getAbsBeat()+ 0.001);
      if (tbeat == beat) {
         endbeat = j;
      } else {
         break;
      }
   }

   for (j=ind; j>=0; j--) {
      tbeat = (int)(infile[index[j]].getAbsBeat()+ 0.001);
      if (tbeat == beat) {
         startbeat = j;
      } else {
         break;
      }
   }

   // don't beam quarter notes or larger
   if (startbeat == endbeat) {
      return;
   }

   // know that some beaming action necessary

   Array<double> durs;
   durs.setSize(0);
   durs.setSize(endbeat-startbeat+1);
   durs.setAll(0.0);

   for (j=0; j<durs.getSize(); j++) {
      durs[j] =  Convert::kernToDuration(infile[index[j+startbeat]][voice]);
   }

   // beam on a case-by-case basis
   double hpos1 = 0.0;
   double hpos2 = 0.0;
   double hpos3 = 0.0;
   double vpos1 = 0.0;
   double vpos2 = 0.0;
   double vpos3 = 0.0;

   //////////

   int pitch1;
   int pitch2;
   int pitch3;
   int dir = 0;

   // beam two eighth notes together

   if ((durs.getSize() == 2) && (durs[0] == 0.5) && (durs[1] == 0.5)
         && fraction < 0.001) {
      hpos1 = getHpos(infile, index[startbeat]);
      hpos2 = getHpos(infile, index[endbeat]);
      pitch1 = Convert::kernToBase40(infile[index[startbeat]][voice]);
      pitch2 = Convert::kernToBase40(infile[index[endbeat]][voice]);
      switch (voice) {
         case 0: case 2:
            if (voice == 0) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
            }
            dir = 21;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos2, 20)
                 << " " 
                 << getBeamEnd(vpos1, vpos2, 20)
                 << " " << hpos2 << " " << dir;
            if (offsets[voice][startbeat] != 0.0) {
                cout << " 0 0 0 0 0 0 " << offsets[voice][startbeat];
                if (offsets[voice][endbeat] != 0.0) {
                   cout << " " << offsets[voice][endbeat];
                }
            } else if (offsets[voice][endbeat] != 0.0) {
                cout << " 0 0 0 0 0 0 0 " << offsets[voice][endbeat];
            }
            cout << "\n";
            break;
         case 1: case 3:
            if (voice == 1) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
            }
            dir = 11;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos2, 10)
                 << " " 
                 << getBeamEnd(vpos1, vpos2, 10)
                 << " " << hpos2 << " " << dir;
            if (offsets[voice][startbeat] != 0.0) {
                cout << " 0 0 0 0 0 0 " << offsets[voice][startbeat];
                if (offsets[voice][endbeat] != 0.0) {
                   cout << " " << offsets[voice][endbeat];
                }
            } else if (offsets[voice][endbeat] != 0.0) {
                cout << " 0 0 0 0 0 0 0 " << offsets[voice][endbeat];
            }
            cout << "\n";
            break;
      }
   }


   //////////

   // beam two sixteenth notes together on the half beat
   else if ((durs.getSize() == 2) && (durs[0] == 0.25) && (durs[1] == 0.25)
         && fraction == 0.5) {
      hpos1 = getHpos(infile, index[startbeat]);
      hpos2 = getHpos(infile, index[endbeat]);
      pitch1 = Convert::kernToBase40(infile[index[startbeat]][voice]);
      pitch2 = Convert::kernToBase40(infile[index[endbeat]][voice]);
      switch (voice) {
         case 0: case 2:
            if (voice == 0) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
            }
            dir = 22;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos2, 20)
                 << " " 
                 << getBeamEnd(vpos1, vpos2, 20)
                 << " " << hpos2 << " " << dir;
            if (offsets[voice][startbeat] > 0.0) {
                cout << " 0 0 0 0 0 0 " << offsets[voice][startbeat];
                if (offsets[voice][endbeat] > 0.0) {
                   cout << " " << offsets[voice][endbeat];
                }
            } else if (offsets[voice][endbeat] > 0.0) {
                cout << " 0 0 0 0 0 0 0 " << offsets[voice][endbeat];
            }
            cout << "\n";
            break;
         case 1: case 3:
            if (voice == 1) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
            }
            dir = 12;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos2, 10)
                 << " " 
                 << getBeamEnd(vpos1, vpos2, 10)
                 << " " << hpos2 << " " << dir;
            if (offsets[voice][startbeat] > 0.0) {
                cout << " 0 0 0 0 0 0 " << offsets[voice][startbeat];
                if (offsets[voice][endbeat] > 0.0) {
                   cout << " " << offsets[voice][endbeat];
                }
            } else if (offsets[voice][endbeat] > 0.0) {
                cout << " 0 0 0 0 0 0 0 " << offsets[voice][endbeat];
            }
            cout << "\n";
            break;
      }
   }

   //////////

   // beam two sixteenth notes starting on the beat (long note or rest
   // on the second half of the beat
   else if (((durs.getSize() == 2) && (durs[0] == 0.25) && (durs[1] == 0.25)
         && fraction == 0.0) ||
         ((durs.getSize() == 3) && (durs[0] == 0.25) && (durs[1] == 0.25)
         && (durs[2] >= 1.0) && fraction == 0.0) 
      ) {

      hpos1 = getHpos(infile, index[startbeat]);
      hpos2 = getHpos(infile, index[startbeat+1]);
      pitch1 = Convert::kernToBase40(infile[index[startbeat]][voice]);
      pitch2 = Convert::kernToBase40(infile[index[startbeat+1]][voice]);
      switch (voice) {
         case 0: case 2:
            if (voice == 0) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
            }
            dir = 22;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos2, 20)
                 << " " 
                 << getBeamEnd(vpos1, vpos2, 20)
                 << " " << hpos2 << " " << dir;
            if (offsets[voice][startbeat] > 0.0) {
                cout << " 0 0 0 0 0 0 " << offsets[voice][startbeat];
                if (offsets[voice][startbeat+1] > 0.0) {
                   cout << " " << offsets[voice][startbeat+1];
                }
            } else if (offsets[voice][startbeat+1] > 0.0) {
                cout << " 0 0 0 0 0 0 0 " << offsets[voice][startbeat+1];
            }
            cout << "\n";
            break;
         case 1: case 3:
            if (voice == 1) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
            }
            dir = 12;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos2, 10)
                 << " " 
                 << getBeamEnd(vpos1, vpos2, 10)
                 << " " << hpos2 << " " << dir;
            if (offsets[voice][startbeat] > 0.0) {
                cout << " 0 0 0 0 0 0 " << offsets[voice][startbeat];
                if (offsets[voice][startbeat+1] > 0.0) {
                   cout << " " << offsets[voice][startbeat+1];
                }
            } else if (offsets[voice][startbeat+1] > 0.0) {
                cout << " 0 0 0 0 0 0 0 " << offsets[voice][startbeat+1];
            }
            cout << "\n";
            break;
      }
   }

   //////////

   // beam one eighth and two sixteenth notes together 
   else if ((durs.getSize() == 3) && (durs[0] == 0.5) && 
         (durs[1] == 0.25) && (durs[2] == 0.25) && fraction == 0.0) {
      hpos1 = getHpos(infile, index[startbeat]);
      hpos2 = getHpos(infile, index[startbeat+1]);
      hpos3 = getHpos(infile, index[endbeat]);
      pitch1 = Convert::kernToBase40(infile[index[startbeat]][voice]);
      pitch2 = Convert::kernToBase40(infile[index[startbeat+1]][voice]);
      pitch3 = Convert::kernToBase40(infile[index[endbeat]][voice]);
      switch (voice) {
         case 0: case 2:
            if (voice == 0) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
               vpos3 = Convert::base40ToScoreVPos(pitch3, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
               vpos3 = Convert::base40ToScoreVPos(pitch3, 0);
            }
            dir = 21;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos3, 20)
                 << " " 
                 << getBeamEnd(vpos1, vpos3, 20) - 0.5
                 << " " << hpos3 << " " << dir 
                 << " 0 0 " << 11 << " " << hpos2 << " " << hpos3;
            if (offsets[voice][startbeat] > 0.0) {
                cout << " 0 " << offsets[voice][startbeat];
                if (offsets[voice][endbeat] > 0.0) {
                   cout << " " << offsets[voice][endbeat];
                }
            } else if (offsets[voice][endbeat] > 0.0) {
                cout << " 0 0 " << offsets[voice][endbeat];
            }
            cout << "\n";
            break;
         case 1: case 3:
            if (voice == 1) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
               vpos3 = Convert::base40ToScoreVPos(pitch3, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
               vpos3 = Convert::base40ToScoreVPos(pitch3, 0);
            }
            dir = 11;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos3, 10)
                 << " " 
                 << getBeamEnd(vpos1, vpos3, 10) + 0.5
                 << " " << hpos3 << " " << dir 
                 << " 0 0 " << 11 << " " << hpos2 << " " << hpos3;
            if (offsets[voice][startbeat] > 0.0) {
                cout << " 0 " << offsets[voice][startbeat];
                if (offsets[voice][endbeat] > 0.0) {
                   cout << " " << offsets[voice][endbeat];
                }
            } else if (offsets[voice][endbeat] > 0.0) {
                cout << " 0 0 " << offsets[voice][endbeat];
            }
            cout << "\n";
            break;
      }
   }

   //////////

   // beam two sixteenths and one eighth note together 
   else if ((durs.getSize() == 3) && (durs[0] == 0.25) && 
         (durs[1] == 0.25) && (durs[2] == 0.5) && fraction == 0.0) {
      hpos1 = getHpos(infile, index[startbeat]);
      hpos2 = getHpos(infile, index[startbeat+1]);
      hpos3 = getHpos(infile, index[endbeat]);
      pitch1 = Convert::kernToBase40(infile[index[startbeat]][voice]);
      pitch2 = Convert::kernToBase40(infile[index[startbeat+1]][voice]);
      pitch3 = Convert::kernToBase40(infile[index[endbeat]][voice]);
      switch (voice) {
         case 0: case 2:
            if (voice == 0) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
               vpos3 = Convert::base40ToScoreVPos(pitch3, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
               vpos3 = Convert::base40ToScoreVPos(pitch3, 0);
            }
            dir = 21;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos3, 20)
                 << " " 
                 << getBeamEnd(vpos1, vpos3, 20) - 0.5
                 << " " << hpos3 << " " << dir 
                 << " 0 0 " << 11 << " " << hpos1 << " " << hpos2;
            if (offsets[voice][startbeat] > 0.0) {
                cout << " 0 " << offsets[voice][startbeat];
                if (offsets[voice][endbeat] > 0.0) {
                   cout << " " << offsets[voice][endbeat];
                }
            } else if (offsets[voice][endbeat] > 0.0) {
                cout << " 0 0 " << offsets[voice][endbeat];
            }
            cout << "\n";
            break;
         case 1: case 3:
            if (voice == 1) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
               vpos3 = Convert::base40ToScoreVPos(pitch3, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
               vpos3 = Convert::base40ToScoreVPos(pitch3, 0);
            }
            dir = 11;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos3, 10)
                 << " " 
                 << getBeamEnd(vpos1, vpos3, 10) + 0.5
                 << " " << hpos3 << " " << dir 
                 << " 0 0 " << 11 << " " << hpos1 << " " << hpos2;
            if (offsets[voice][startbeat] > 0.0) {
                cout << " 0 " << offsets[voice][startbeat];
                if (offsets[voice][endbeat] > 0.0) {
                   cout << " " << offsets[voice][endbeat];
                }
            } else if (offsets[voice][endbeat] > 0.0) {
                cout << " 0 0 " << offsets[voice][endbeat];
            }
            cout << "\n";
            break;
      }
   }

   //////////

   // beam dotted eight and sixteenth
   else if ((durs.getSize() == 2) && (durs[0] == 0.75) && (durs[1] == 0.25)
         && fraction == 0.0) {
      hpos1 = getHpos(infile, index[startbeat]);
      hpos2 = getHpos(infile, index[endbeat]);
      pitch1 = Convert::kernToBase40(infile[index[startbeat]][voice]);
      pitch2 = Convert::kernToBase40(infile[index[endbeat]][voice]);
      switch (voice) {
         case 0: case 2:
            if (voice == 0) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
            }
            dir = 21;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos2, 20)
                 << " " 
                 << getBeamEnd(vpos1, vpos2, 20)
                 << " " << hpos2 << " " << dir << " 0 0 "
                 << 11 << " 0 " << hpos2;
            if (offsets[voice][startbeat] > 0.0) {
                cout << " 0 " << offsets[voice][startbeat];
                if (offsets[voice][endbeat] > 0.0) {
                   cout << " " << offsets[voice][endbeat];
                }
            } else if (offsets[voice][endbeat] > 0.0) {
                cout << " 0 0 " << offsets[voice][endbeat];
            }
            cout << "\n";
            break;
         case 1: case 3:
            if (voice == 1) {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 1);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 1);
            } else {
               vpos1 = Convert::base40ToScoreVPos(pitch1, 0);
               vpos2 = Convert::base40ToScoreVPos(pitch2, 0);
            }
            dir = 11;
            cout << "6 " << voice+1 << " " << hpos1 << " "
                 << getBeamStart(vpos1, vpos2, 10)
                 << " " 
                 << getBeamEnd(vpos1, vpos2, 10)
                 << " " << hpos2 << " " << dir << " 0 0 "
                 << 11 << " 0 " << hpos2;
            if (offsets[voice][startbeat] > 0.0) {
                cout << " 0 " << offsets[voice][startbeat];
                if (offsets[voice][endbeat] > 0.0) {
                   cout << " " << offsets[voice][endbeat];
                }
            } else if (offsets[voice][endbeat] > 0.0) {
                cout << " 0 0 " << offsets[voice][endbeat];
            }
            cout << "\n";
            break;
      }
   }

}



//////////////////////////////
//
// getHpos -- return the horizontal postion of an item in the score
//

double getHpos(HumdrumFile& infile, int line) {
   double output;
   output = infile[line].getAbsBeat() / infile.getTotalDuration() * 190 + 10;
   output = ((int)(output*100.0 + 0.5))/100.0;
   return output;
}



//////////////////////////////
//
// getBeamStart --
//

double getBeamStart(double vpos1, double vpos2, int dir) {

   double upbeam1[23][23] = {{0, 0, 0, 0, 0, 0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{0, 0, 0, 0, 0, 0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{0, 0, 0, 0, 0, 0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{0, 0, 0, 0, 0, 0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{1, 1, 1, 1, 1, 1, 1.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{2, 2, 2, 2, 2, 2, 2, 2.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{2, 2, 2, 2, 3, 3, 3, 3, 3.5, 3.5, 4.5, 5.5, 6.5, 7.5, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{3, 3, 3, 3, 3, 4, 4, 4, 4, 4.5, 4.5, 5.5, 6.5, 7.5, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5.5, 5.5, 6.5, 7.5, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6.5, 6.5, 7.5, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7.5, 7.5, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{7, 7, 7, 7, 7, 7, 7, 7, 7, 7.5, 7.5, 7.5, 7.5, 7.75, 7.75, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{8.5, 8.5, 8.5, 8.5, 8.5, 8.5, 8.5, 8.5, 8.5, 8.5, 8, 8, 8, 8, 8.25, 8.25, 9, 10, 11, 12, 13, 14, 15},
	{9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 8.5, 8.5, 9, 8.5, 9, 9, 10, 11, 12, 13, 14, 15},
	{9.5, 9.5, 9.5, 9.5, 9.5, 9.5, 9.5, 9.5, 9.5, 9.5, 9.5, 9.5, 9.5, 9.5, 9.5, 9.5, 10, 10, 11, 12, 13, 14, 15},
	{10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 10.5, 11, 11, 12, 13, 14, 15},
	{11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 11.5, 12, 12, 13, 14, 15},
	{12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 12.5, 13, 13, 14, 15},
	{13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 13.5, 14, 14, 15},
	{14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 15, 15},
	{15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 15.5, 16},
	{16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5, 16.5},
	{17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5, 17.5}};

   double downbeam1[23][23] = {{-1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5},
	{-0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5},
	{0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5},
	{1, 1, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5},
	{1, 1.5, 1.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5},
	{1, 1.5, 2.5, 2.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5},
	{1, 1.5, 2.5, 3.5, 3.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5},
	{1, 1.5, 2.5, 3.5, 4.5, 4.5, 5.5, 5, 5.5, 5.5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 5.5, 6, 6, 6, 6, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6, 6.5, 6.5, 6.5, 6.5, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 6.5, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 8, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 8, 9, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 8, 9, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 8, 9, 10, 11, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 14, 14, 14, 14, 14},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 14, 14, 14, 14, 14},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 14, 14, 14, 14, 14},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 14, 14, 14, 14, 14},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 14, 14, 14, 14, 14},
	{1, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 14, 14, 14, 14, 14}};

   if (vpos1 < -3 || vpos2 < -3 || vpos1 > 22 || vpos2 > 22) {
      return vpos1;
   }

   switch (dir) {
      case 10: return upbeam1[(int)vpos1+3][(int)vpos2+3];
      case 20: return downbeam1[(int)vpos1+3][(int)vpos2+3];
   }

   return vpos1;
}



//////////////////////////////
//
// getBeamEnd --
//


double getBeamEnd(double vpos1, double vpos2, int dir) {

   double upbeam2[23][23] = {{0, 0, 0, 0, 1, 2, 2, 3, 4, 5, 6, 7, 8.5, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{0, 0, 0, 0, 1, 2, 2, 3, 4, 5, 6, 7, 8.5, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{0, 0, 0, 0, 1, 2, 2, 3, 4, 5, 6, 7, 8.5, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{0, 0, 0, 0, 1, 2, 2, 3, 4, 5, 6, 7, 8.5, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{0, 0, 0, 0, 1, 2, 3, 3, 4, 5, 6, 7, 8.5, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{0.5, 0.5, 0.5, 0.5, 1, 2, 3, 4, 4, 5, 6, 7, 8.5, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{0.5, 0.5, 0.5, 0.5, 1.5, 2, 3, 4, 5, 5, 6, 7, 8.5, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{1.5, 1.5, 1.5, 1.5, 1.5, 2.5, 3, 4, 5, 6, 6, 7, 8.5, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 3.5, 4, 5, 6, 7, 7, 8.5, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 4.5, 5, 6, 7, 7.5, 8.5, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 5.5, 6, 7, 7.5, 8, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 6.5, 7, 7.5, 8, 8.5, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 7.5, 7.5, 8, 8.5, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{7.5, 7.5, 7.5, 7.5, 7.5, 7.5, 7.5, 7.5, 7.5, 7.5, 7.5, 7.75, 8, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{7.75, 7.75, 7.75, 7.75, 7.75, 7.75, 7.75, 7.75, 7.75, 7.75, 7.75, 7.75, 8.25, 8.5, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{8.25, 8.25, 8.25, 8.25, 8.25, 8.25, 8.25, 8.25, 8.25, 8.25, 8.25, 8.25, 8.25, 9, 9.5, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10.5, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11.5, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12.5, 13.5, 14.5, 15.5, 16.5, 17.5},
	{12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13.5, 14.5, 15.5, 16.5, 17.5},
	{13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14.5, 15.5, 16.5, 17.5},
	{14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15.5, 16.5, 17.5},
	{15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16.5, 17.5}};

   double downbeam2[23][23] = {{-1.5, -0.5, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1},
	{-1.5, -0.5, 0.5, 1, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5},
	{-1.5, -0.5, 0.5, 1.5, 1.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 2.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 3.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 4.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5, 5.5},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 5.5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 6, 6, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5, 6.5},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 6.5, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6, 6.5, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 6, 6.5, 7, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 6.5, 7, 8, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 7, 7, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 7, 8, 8, 9, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 7, 8, 9, 9, 10, 11, 12, 12, 13, 13, 13, 13, 13, 13},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 7, 8, 9, 10, 10, 11, 12, 13, 13, 13, 13, 13, 13, 13},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 14, 14, 14, 14},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 14, 14, 14, 14},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 14, 14, 14, 14},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 14, 14, 14, 14},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 14, 14, 14, 14},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 14, 14, 14, 14},
	{-1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5, 5.5, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 14, 14, 14, 14}};

   if (vpos1 < -3 || vpos2 < -3 || vpos1 > 22 || vpos2 > 22) {
      return vpos2;
   }

   switch (dir) {
      case 10: return upbeam2[(int)vpos1+3][(int)vpos2+3];
      case 20: return downbeam2[(int)vpos1+3][(int)vpos2+3];
   }

   return vpos2;
}



